package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.properties.ApplicationProperties;
import at.ac.uibk.gitsearch.properties.ApplicationProperties.RegisteredConnector;
import at.ac.uibk.gitsearch.service.PluginManagementService.ConnectorConfigWrapper.ActionWrapper;
import com.fasterxml.jackson.jaxrs.json.JacksonJsonProvider;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Timer;
import java.util.TimerTask;
import javax.annotation.PostConstruct;
import javax.el.ELContext;
import javax.el.ExpressionFactory;
import javax.el.StandardELContext;
import javax.el.ValueExpression;
import javax.ws.rs.ProcessingException;
import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.Invocation.Builder;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.MediaType;
import org.apache.commons.lang3.StringUtils;
import org.codeability.sharing.plugins.api.SharingPluginConfig;
import org.codeability.sharing.plugins.api.SharingPluginConfig.Action;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO.PluginActionInfo;
import org.glassfish.jersey.client.ClientConfig;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

/**
 * management services for plugins.
 *
 * @author Michael Breu
 *
 */
@Service
public class PluginManagementService {

    private static final String API_BASE_URL_PARAM = "apiBaseUrl";
    private static final String INSTALLATION_NAME_PARAM = "installationName";

    private final Map<RegisteredConnector, Exception> configFailures = new HashMap<>();

    private final class PluginCheckTimer extends TimerTask {

        private final RegisteredConnector registeredConnector;

        private PluginCheckTimer(RegisteredConnector registeredConnector) {
            this.registeredConnector = registeredConnector;
        }

        @Override
        public void run() {
            synchronized (configFailures) {
                ClientConfig restClientConfig = new ClientConfig();
                restClientConfig.register(JacksonJsonProvider.class);
                Client client = ClientBuilder.newClient(restClientConfig);

                WebTarget target = client
                    .target(registeredConnector.getUrl())
                    //						.queryParam(ACCESS_TOKEN_PARAM, registeredPlugin.getAccessToken())
                    .queryParam(API_BASE_URL_PARAM, applicationProperties.getRegisteredConnectorsCallBackURL())
                    .queryParam(INSTALLATION_NAME_PARAM, applicationProperties.getInstallationName());
                SharingPluginConfig config = null;
                try {
                    Builder request = target.request().accept(MediaType.APPLICATION_JSON);

                    if (!StringUtils.isEmpty(registeredConnector.getAccessToken())) request =
                        request.header("Authorization", "Bearer " + registeredConnector.getAccessToken());
                    config = request.get(SharingPluginConfig.class);
                    registeredPluginConfigs.put(config.pluginName, new ConnectorConfigWrapper(config, registeredConnector.getUrl()));
                    configFailures.remove(registeredConnector);
                } catch (ProcessingException ce) {
                    log.warn("Cannot connect to connector at {}: {}", registeredConnector.getUrl(), ce.getMessage());
                    configFailures.put(registeredConnector, ce);
                } catch (Exception e) {
                    log.warn("Cannot (re-)load connector at {}", registeredConnector, e.getMessage());
                    configFailures.put(registeredConnector, e);
                    final Iterator<Entry<String, ConnectorConfigWrapper>> pluginIterator = registeredPluginConfigs.entrySet().iterator();
                    while (pluginIterator.hasNext()) {
                        final Entry<String, ConnectorConfigWrapper> nextEntry = pluginIterator.next();
                        if (nextEntry.getValue().getConfigURL().equals(registeredConnector.getUrl())) {
                            pluginIterator.remove();
                            break;
                        }
                    }
                }
            }
        }
    }

    /**
     * this is just a wrapper, to provide a more efficient access to the original
     * config.
     *
     * @author Michael Breu
     *
     */
    public static class ConnectorConfigWrapper {

        public static class ActionWrapper {

            protected SharingPluginConfig.Action originalAction;
            protected ValueExpression filterExpression;
            protected PluginActionInfo pluginActionInfo;

            String absoluteActionTransferURL;

            public ActionWrapper(Action action, ConnectorConfigWrapper plugin) {
                super();
                this.originalAction = action;
                pluginActionInfo = new PluginActionInfo(plugin.getPluginName(), action.actionId, action.actionName);

                try {
                    factory.createValueExpression(factory, Boolean.class);
                    ELContext context = new StandardELContext(factory);
                    filterExpression =
                        factory.createValueExpression(context, "${" + originalAction.filterELExpression + "}", Boolean.class);
                } catch (Exception e) {
                    log.error(
                        "Cannot parse elExpression {} for pluginAction {}:{} ",
                        originalAction.filterELExpression,
                        plugin.getPluginName(),
                        action.actionId
                    );
                    filterExpression = null;
                }

                try {
                    URI actionURI = new URI(action.actionTransferURL);
                    if (!actionURI.isAbsolute()) {
                        this.absoluteActionTransferURL = new URI(plugin.getConfigURL()).resolve(actionURI).toASCIIString();
                    }
                } catch (URISyntaxException e) {
                    log.error(
                        "Cannot asolutize url {} from action {} via {}",
                        action.actionTransferURL,
                        plugin.getConfigURL(),
                        plugin.getPluginName(),
                        e
                    );
                }
            }

            /**
             * @return the absoluteImportServiceURL
             */
            public String getAbsoluteActionTransferURL() {
                return absoluteActionTransferURL;
            }

            /**
             * @return the originalAction
             */
            public SharingPluginConfig.Action getOriginalAction() {
                return originalAction;
            }

            public String getActionId() {
                return originalAction.actionId;
            }

            public String getActionName() {
                return originalAction.actionName;
            }

            public PluginActionInfo getPluginActionInfo() {
                return pluginActionInfo;
            }

            protected static ExpressionFactory factory = ExpressionFactory.newInstance();

            public boolean isApplicable(final SearchResultDTO hit) {
                if (filterExpression == null) return false;
                try {
                    StandardELContext context = new StandardELContext(factory);
                    context.getELResolver().setValue(context, null, "metadata", hit.getMetadata());
                    return (boolean) filterExpression.getValue(context);
                } catch (Exception e) {
                    log.error(
                        "Cannot evaluate elExpression {} for pluginAction {}:{} ",
                        originalAction.filterELExpression,
                        originalAction.actionId,
                        e
                    );
                    return false;
                }
            }
        }

        protected SharingPluginConfig originalConfig;
        protected Map<String, ActionWrapper> actions = new HashMap<>();
        protected String configURL;

        public ConnectorConfigWrapper(SharingPluginConfig originalConfig, String configURL) {
            super();
            this.originalConfig = originalConfig;
            this.configURL = configURL;
            actions = new HashMap<>();
            for (Action action : originalConfig.actions) {
                actions.put(action.actionId, new ActionWrapper(action, this));
            }
        }

        /**
         * @return the baseURL
         */
        public String getConfigURL() {
            return configURL;
        }

        public String getPluginName() {
            return originalConfig.pluginName;
        }

        public Map<String, ActionWrapper> getActions() {
            return actions;
        }
    }

    @Autowired
    private ApplicationProperties applicationProperties;

    private static Map<String, ConnectorConfigWrapper> registeredPluginConfigs = new HashMap<>();

    private static final Logger log = LoggerFactory.getLogger(PluginManagementService.class);

    private static final long PLUGIN_RECHECK_TIMEOUT = 5 * 60 * 1000L; // every 5 Minutes

    /**
     * triggers registration of all plugins after startup.
     */
    @PostConstruct
    public void registerPlugins() {
        List<RegisteredConnector> registeredPlugins = applicationProperties.getRegisteredConnectors();

        if (registeredPlugins == null) {
            log.warn("No Plugins registered?");
            return;
        }

        for (RegisteredConnector registeredPlugin : registeredPlugins) {
            Timer check = new Timer(registeredPlugin + "-CheckTimer");
            check.schedule(new PluginCheckTimer(registeredPlugin), 0L, PLUGIN_RECHECK_TIMEOUT);
        }
    }

    public String getRedirectURL(String pluginName, String actionId) {
        final ConnectorConfigWrapper sharingPluginConfig = registeredPluginConfigs.get(pluginName);
        if (sharingPluginConfig == null) return null;
        final ActionWrapper actionWrapper = sharingPluginConfig.actions.get(actionId);
        if (actionWrapper == null) return null;
        return actionWrapper.getAbsoluteActionTransferURL();
    }

    public Collection<ConnectorConfigWrapper> getRegisteredPluginConfigs() {
        return registeredPluginConfigs.values();
    }

    /**
     * @return the configFailures
     */
    public Map<RegisteredConnector, Exception> getConfigFailures() {
        return Collections.unmodifiableMap(configFailures);
    }
}
